package influx.adapter;
import java.time.Instant;
import java.util.Iterator;
import java.util.concurrent.TimeUnit;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.influxdb.BatchOptions;
import org.influxdb.InfluxDB;
import org.influxdb.InfluxDBFactory;
import org.influxdb.dto.Point;
import org.influxdb.dto.Point.Builder;
import org.influxdb.dto.Query;
import org.influxdb.dto.QueryResult;

public class DBConnector {

	private InfluxDB influxDB;
	private static DBConnector INSTANCE;

	private DBConnector() {
	}

	public static DBConnector getInstance() {
		if (INSTANCE == null) {
			INSTANCE = new DBConnector();
		}
		return INSTANCE;
	}

	/*
	 * DB operations
	 * 
	 */
	
	public void startDB(String serverURL, String username, String password, String databaseName, String duration) {
		influxDB = InfluxDBFactory.connect(serverURL, username, password);
		influxDB.setDatabase(databaseName);
		influxDB.enableBatch(BatchOptions.DEFAULTS);
	}
	
	public void startConnection(String serverURL, String username, String password, String databaseName, String duration) {
		influxDB = InfluxDBFactory.connect(serverURL, username, password);

		// this may not be intended
		influxDB.query(new Query("DROP DATABASE " + databaseName));
		influxDB.query(new Query("CREATE DATABASE " + databaseName));
		influxDB.setDatabase(databaseName);

		// may be moved to the configuration model
		String retentionPolicyName = "one_day_only";
		influxDB.query(new Query("CREATE RETENTION POLICY " + retentionPolicyName + " ON " + databaseName
				+ " DURATION " + duration + " REPLICATION 1 DEFAULT"));
		influxDB.setRetentionPolicy(retentionPolicyName);

		influxDB.enableBatch(BatchOptions.DEFAULTS);

	}
	
	public void flush() {
		influxDB.flush();
	}
	
	public void closeConnection() {
		influxDB.close();
	}

	/*
	 * DB update operations
	 * 
	 */
	
	public void writeToDB(String measurement, String tag, Integer field) {
		influxDB.write(Point.measurement(measurement).time(System.currentTimeMillis(), TimeUnit.MILLISECONDS)
				.tag("object", tag).addField("value", field).build());
	}
	
	public void writeToDB(String measurement, String tag, String field) {
		influxDB.write(Point.measurement(measurement).time(System.currentTimeMillis(), TimeUnit.MILLISECONDS)
				.tag("object", tag).addField("value", field).build());
	}
	
	// mapping Property2TSDB Strategy STARTS HERE
	public void writeToDB(String measurement, String tag, String field, String precision) {
		switch(precision) {
		case "ms":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().toEpochMilli(), TimeUnit.MILLISECONDS).build());
		case "s":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().getEpochSecond(), TimeUnit.SECONDS).build());
		case "us":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].substring(0, 6)), TimeUnit.MICROSECONDS).build());
		default:
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].split("Z")[0]), TimeUnit.NANOSECONDS).build());	
		}
	}
	
	public void writeToDB(String measurement, String tag, int field, String precision) {
		switch(precision) {
		case "ms":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().toEpochMilli(), TimeUnit.MILLISECONDS).build());
		case "s":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().getEpochSecond(), TimeUnit.SECONDS).build());
		case "us":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].substring(0, 6)), TimeUnit.MICROSECONDS).build());
		default:
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].split("Z")[0]), TimeUnit.NANOSECONDS).build());	
		}
	}
	
	public void writeToDB(String measurement, String tag, boolean field, String precision) {
		switch(precision) {
		case "ms":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().toEpochMilli(), TimeUnit.MILLISECONDS).build());
		case "s":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().getEpochSecond(), TimeUnit.SECONDS).build());
		case "us":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].substring(0, 6)), TimeUnit.MICROSECONDS).build());
		default:
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].split("Z")[0]), TimeUnit.NANOSECONDS).build());	
		}
	}
	
	public void writeToDB(String measurement, String tag, float field, String precision) {
		switch(precision) {
		case "ms":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().toEpochMilli(), TimeUnit.MILLISECONDS).build());
		case "s":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Instant.now().getEpochSecond(), TimeUnit.SECONDS).build());
		case "us":
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].substring(0, 6)), TimeUnit.MICROSECONDS).build());
		default:
			influxDB.write(Point.measurement(measurement).tag("object", tag).addField("value", field).time(Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].split("Z")[0]), TimeUnit.NANOSECONDS).build());	
		}
	}
	
	
	// mapping Object2TSDB Strategy STARTS HERE
	public void writeToDB(EObject object, String precision) {
		Builder b = Point.measurement("obj" + System.identityHashCode(object));
		EList<EStructuralFeature> attributes = object.eClass().getEAllStructuralFeatures();
		
		Iterator<EStructuralFeature> i = attributes.iterator();
		while(i.hasNext()) {
			EStructuralFeature att = i.next();
			if(att.isDerived()) continue;
			Object value = object.eGet(att);
			if(value != null) {
				switch(att.getEType().getName()) {
				case "EInt":
					b.addField(att.getName(), (Integer) object.eGet(att));
				case "EFloat":
					b.addField(att.getName(), (Float) object.eGet(att));
				case "EBoolean":
					b.addField(att.getName(), (Boolean) object.eGet(att));
				default:
					b.addField(att.getName(), object.eGet(att)+"");	
				}
			}
			else{
				b.addField(att.getName(), "undefined");
			}
		}
		Long timestamp;
		TimeUnit prec;
		
		switch(precision) {
		case "ms":
			timestamp=Instant.now().toEpochMilli();
			prec=TimeUnit.MILLISECONDS;
		case "s":
			timestamp = Instant.now().getEpochSecond();
			prec= TimeUnit.SECONDS;
		case "us":
			timestamp = Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].substring(0, 6));
			prec= TimeUnit.SECONDS;
		default:
			timestamp=Long.parseLong(Instant.now().getEpochSecond()+Instant.now().toString().split("\\.")[1].split("Z")[0]);
			prec=TimeUnit.NANOSECONDS;	
		}
		
		influxDB.write(b.time(timestamp,prec).build());	
	}
	

	/*
	 * DB query operations
	 * 
	 */
	
	// queries for Property2TSDB Strategy
	public String queryDB(String query) {
		QueryResult queryResult = influxDB.query(new Query(query));
		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(1).toString();
		}
	}
	
	public String getValueAt(String time, String object, String feature) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select value from " + feature + " where object = '"+ object +"' and time = '"+ time+"'"));
		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(1).toString();
		}
	}
	
	public String getValueBetween(String time1, String time2, String object, String feature) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select value from " + feature + " where object = '"+ object +"' and time >= " + time1 + " and time <= " + time2));
		return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(1).toString();
	}
	
	public String getTimePointsForValue(String object, String feature, String value) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select time, value from " + feature + " where object = '"+ object +"' and value = " + value));
		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(0).toString();
		}
	}
	
	public String getTimePointsForValueBetween(String object, String feature, String value1,String value2) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select time, value from " + feature + " where object = '"+ object +"' and value >= " + value1+ " and value <= " + value2));
		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(0).toString();
		}
	}
	
	// queries for object mapping
	public String getValueAt(String time, EObject object, String feature) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select " + feature + " from " + "obj" + System.identityHashCode(object) + " where time = '"+ time+"'"));

		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(1).toString();
		}

	}
	public String getValueBetween(String time1, String time2, EObject object, String feature) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select " + feature + " from " + "obj" + System.identityHashCode(object) + " where time >= " + time1 + " and time <= " + time2));

		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(1).toString();
		}
	}
	
	public String getTimePointsForValue(EObject object, String feature, String value) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select time, "+ feature +"  from " + "obj" + System.identityHashCode(object) + " where " + feature + " = " + value));
		
		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(0).toString();
		}
	}
	public String getTimePointsForValueBetween(EObject object, String feature, String value1,String value2) {
		QueryResult queryResult = influxDB.query(new Query(
				"Select time, "+ feature +"  from " + "obj" + System.identityHashCode(object) + " where " + feature + " >= " + value1 + "and "+ feature + " <= " + value2));
		
		if(queryResult.getResults().get(0).getSeries()==null) {
			return "";
		}else {
			return queryResult.getResults().get(0).getSeries().get(0).getValues().get(0).get(0).toString();
		}
	}
	
	

}
