package modelSimCase;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;

import org.junit.Test;

import influx.adapter.DBConnector;
import prodSimObj.Component;
import prodSimObj.DatabaseConfiguration;
import prodSimObj.ProdSimObjFactory;
import prodSimObj.management.ModelDeserializer;

public class ModelObjects2TS {
	
    private static final String DB_folder = "C:\\Users\\Sabine\\.influxdb\\data\\system1";
	private static final int numberOfRuns = 5;
    
	private ArrayList<String[]> results = new ArrayList<String[]>();
	private WriteCsvFileExample csvExample = new WriteCsvFileExample();
    
	@Test
	public void runBenchmark() throws IOException, InterruptedException {
		System.out.println("Benchmark Start");
		results.add(new String[] {"FullObjects2TimeSeries"});
		results.add(new String[] { "Components", "Runs", "SimTime", "DB size", "MeanTemp", "TimePoints4Value", "GetValueAtTimePoint" });

		// start different runs 
		for(int i = 0; i < numberOfRuns; i++) { 
			results.add(new String[] {"Run " + i});
			testMapFullObj();
		}
		
		// store results in CSV file 
		convertToCSV(results);
		
		System.out.println("Benchmark End");
	}
	
	public void testMapFullObj() throws InterruptedException, IOException {

		// different scenarios to consider 
		//final int components[]=  {10, 100, 1000, 2000, 3000, 4000,5000};
		//final int runs[] =       {10, 100, 1000, 2000, 3000, 4000,5000};
		
		// just to test if the DB connection and evaluation works
		final int components[]=  {10, 100, 3000};
		final int runs[] =       {10, 100, 3000};
		
		// run small to large combinations 
		for(int i = 0; i < components.length; i++) {
			   int component = components[i];
			   int run = runs[i];	   
			   runModelSim(component, run);
		}
	}
    
	public void runModelSim(int components, int runs) throws InterruptedException, IOException {
		
		System.out.println("RUN with: " + components + " " + runs);

		DBConnector db = setupDB();
		
		prodSimObj.System sys = createModel4Sim(components);
		
		long timeElapsedSim = runSimulations(runs, sys);
		
		db.flush();
		
		long timeElapsedQuery = measureAggregationQueries(sys);
		
		//warm-up query for measuring the query to get the time points for a value
		//stays here as timestamp is needed for the next measurement - @TODO: needs refactoring
		long start;
		long finish;
		Component comp;
		comp = sys.getComponent().get(2);
		db.getTimePointsForValue(comp, "temperature", "70");
		// measured points for value
	    start = System.nanoTime();
		comp = sys.getComponent().get(9);
		String timestamp = db.getTimePointsForValue(comp, "temperature", "70");
		System.out.println(timestamp);
		finish = System.nanoTime();
		long timeElapsedPointInTime = finish - start;
		
		// measure the query performance of get value for a particular point in time
		long timeElapsedValueAt = measureGetValAtTimePoint(db, sys, timestamp);
		
		// close DB
		db.closeConnection();
		
		//measure database size
		long size = measureDBsize();
		
        results.add(new String[] { components+"", runs+"", timeElapsedSim+"", size+"", timeElapsedQuery+"", timeElapsedValueAt+"", timeElapsedPointInTime+"" });

	}

	private DBConnector setupDB() {
		// Setup Database
		ModelDeserializer md = new ModelDeserializer();
		DatabaseConfiguration dbconfig = (DatabaseConfiguration) md.loadModel("model/dbconfig.xmi");
		DBConnector db = DBConnector.getInstance();
		db.startConnection(dbconfig.getDatabaseconnection().getServerURL(), dbconfig.getDatabaseuser().getUserName(), 
				dbconfig.getDatabaseuser().getPassword(), dbconfig.getDatabaseconnection().getDataBaseName(),
				dbconfig.getRetentionPolicy().getDuration());
		return db;
	}

	private prodSimObj.System createModel4Sim(int components) {
		// create system model
		prodSimObj.System sys = ProdSimObjFactory.eINSTANCE.createSystem();
		sys.setName("system1");
		
		// create as many components as needed
		for (int i = 0; i < components; i++) {
			Component c = ProdSimObjFactory.eINSTANCE.createComponent();
			c.setId(i+"");
			c.setIsActive(false);
			c.setTemperature(70);
			sys.getComponent().add(c);
		}
		return sys;
	}

	private long runSimulations(int runs, prodSimObj.System sys) {
		//measure simulation time
	    long start = System.nanoTime();
		// run the model simulation
		for (int i = 0; i < runs; i++) {
			sys.getComponent().parallelStream().forEach((component) -> {
				component.run();
		    });
		}
		long finish = System.nanoTime();
		long timeElapsedSim = finish - start;
		//System.out.println("Simulation time: " + timeElapsedSim + " in nano-sec");
		return timeElapsedSim;
	}

	private long measureAggregationQueries(prodSimObj.System sys) {
		long start;
		long finish;
		//measure query time
		//warm-up query
		Component comp = sys.getComponent().get(2);
	    comp.getMaxReachedTemperature();
	    // measured query
	    start = System.nanoTime();
	    // query meanTemp
	    comp = sys.getComponent().get(9);
	    float meanTemp = comp.getMeanTemperature();
	    System.out.println(meanTemp);
	    // query maxTemp
		comp = sys.getComponent().get(8);
	    float maxTemp = comp.getMaxReachedTemperature();
	    System.out.println(maxTemp);
	    // query modeTemp 
	    comp = sys.getComponent().get(7);
	    float modeTemp = comp.getModeTemperature();
	    System.out.println(modeTemp);
		
	    finish = System.nanoTime();
		long timeElapsedQuery = finish - start;
		//System.out.println("Query time: " + timeElapsedQuery + " in nano-sec");
		return timeElapsedQuery;
	}

	private long measureGetValAtTimePoint(DBConnector db, prodSimObj.System sys, String timestamp) {
		long start;
		long finish;
		Component comp;
		//warm-up query
		//comp = sys.getComponent().get(2);
	    //db.getValAt(timestamp+"", System.identityHashCode(comp)+"", "temperature");
	    start = System.nanoTime();
		comp = sys.getComponent().get(9);
	    String value = db.getValueAt(timestamp+"", comp, "temperature");
	    System.out.println(value);
		finish = System.nanoTime();
		long timeElapsedValueAt = finish - start;
		//System.out.println("ValueAt time: " + timeElapsedValueAt + " in nano-sec");
		return timeElapsedValueAt;
	}

	private long measureDBsize() throws IOException {
	    Path folder = Paths.get(DB_folder);
	    long size = Files.walk(folder)
	      .filter(p -> p.toFile().isFile())
	      .mapToLong(p -> p.toFile().length())
	      .sum();
		return size;
	}

    public void convertToCSV(ArrayList<String[]> content) throws IOException {

        File csvOutputFile = new File("results/benchmarkResults.csv");
        System.out.println("Results stored in file path: "+csvOutputFile.getAbsolutePath());
        try (PrintWriter pw = new PrintWriter(csvOutputFile)) {
            content.stream()
                .map(csvExample::convertToCSV)
                .forEach(pw::println);
        } catch (FileNotFoundException e) {}
    }
	
}
